@@ -14,17 +14,17 @@ use alloc::vec::Vec;
 
 use num_traits::Unsigned;
 
-/// This trait is implemented for points that exist in a metric space.
+/// This trait is implemented for metrics that form a metric space.
 /// It is primarily used for keys in nearest neighbor searches.
 /// When implementing this trait, you should always choose the smallest unsigned integer that
 /// represents your metric space.
 ///
-/// It is important that all points that implement this trait satisfy
+/// It is important that all metrics that implement this trait satisfy
 /// the [triangle inequality](https://en.wikipedia.org/wiki/Triangle_inequality).
 /// This requirement basically means that the sum of distances that start
 /// at a point A and end at a point B can never be less than the distance
 /// from A to B directly. Note that the metric is required to be an unsigned integer,
-/// as distances can only be positive.
+/// as distances can only be positive and must be fully ordered.
 ///
 /// Floating point numbers can be converted to integer metrics by being interpreted as integers by design,
 /// although some special patterns (like NaN) do not fit into this model. To be interpreted as an unsigned
@@ -35,79 +35,88 @@ use num_traits::Unsigned;
 /// ## Example
 ///
 /// ```
-/// struct A(f64);
+/// struct AbsDiff;
 ///
-/// impl space::MetricPoint for A {
-///     type Metric = u64;
+/// impl space::Metric<f64> for AbsDiff {
+///     type Unit = u64;
 ///
-///     fn distance(&self, other: &Self) -> Self::Metric {
-///         let delta = (self.0 - other.0).abs();
+///     fn distance(&self, &a: &f64, &b: &f64) -> Self::Unit {
+///         let delta = (a - b).abs();
 ///         debug_assert!(!delta.is_nan());
 ///         delta.to_bits()
 ///     }
 /// }
 /// ```
-pub trait MetricPoint {
-    type Metric: Unsigned + Ord + Copy;
+pub trait Metric<P> {
+    type Unit: Unsigned + Ord + Copy;
 
-    fn distance(&self, other: &Self) -> Self::Metric;
+    fn distance(&self, a: &P, b: &P) -> Self::Unit;
 }
 
 /// For k-NN algorithms to return neighbors.
 #[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
-pub struct Neighbor<Metric, Ix = usize> {
+pub struct Neighbor<Unit, Ix = usize> {
     /// Index of the neighbor in the search space.
     pub index: Ix,
     /// The distance of the neighbor from the search feature.
-    pub distance: Metric,
+    pub distance: Unit,
 }
 
 /// Implement this trait on data structures (or wrappers) which perform KNN searches.
-pub trait Knn<P, Ix = usize>
-where
-    P: MetricPoint,
-    Ix: Copy,
-{
-    type KnnIter: IntoIterator<Item = Neighbor<P::Metric, Ix>>;
+pub trait Knn {
+    type Ix: Copy;
+    type Point;
+    type Metric: Metric<Self::Point>;
+    type KnnIter: IntoIterator<
+        Item = Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>,
+    >;
 
     /// Get `num` nearest neighbors of `target`.
     ///
     /// For many KNN search algorithms, the returned neighbors are approximate, and may not
     /// be the actual nearest neighbors.
-    fn knn(&self, query: &P, num: usize) -> Self::KnnIter;
+    fn knn(&self, query: &Self::Point, num: usize) -> Self::KnnIter;
 
     /// Get the nearest neighbor of `target`.
     ///
     /// For many KNN search algorithms, the returned neighbors are approximate, and may not
     /// be the actual nearest neighbors.
-    fn nn(&self, query: &P) -> Option<Neighbor<P::Metric, Ix>> {
+    #[allow(clippy::type_complexity)]
+    fn nn(
+        &self,
+        query: &Self::Point,
+    ) -> Option<Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>> {
         self.knn(query, 1).into_iter().next()
     }
 }
 
-/// This trait gives knn search collections the ability to give the nearest neighbor keys back.
+/// This trait gives knn search collections the ability to give the nearest neighbor points back.
 ///
 /// This is not the final API. Eventually, the iterator type will be chosen by the collection,
 /// but for now it is a [`Vec`] until Rust stabilizes GATs.
 #[cfg(feature = "alloc")]
-pub trait KnnPoints<P, Ix = usize>: Knn<P, Ix>
-where
-    P: MetricPoint,
-    Ix: Copy,
-{
+pub trait KnnPoints: Knn {
     /// Get a point using a neighbor index returned by [`Knn::knn`] or [`Knn::nn`].
     ///
     /// This should only be used directly after one of the mentioned methods are called to retrieve
     /// a point associated with a neighbor, and will panic if the index is incorrect due to
     /// mutating the data structure thereafter. The index is only valid up until the next mutation.
-    fn get_point(&self, index: Ix) -> &'_ P;
+    fn get_point(&self, index: Self::Ix) -> &'_ Self::Point;
 
     /// Get `num` nearest neighbor points of `target`.
     ///
     /// For many KNN search algorithms, the returned neighbors are approximate, and may not
     /// be the actual nearest neighbors.
-    fn knn_points(&self, query: &P, num: usize) -> Vec<(Neighbor<P::Metric, Ix>, &'_ P)> {
+    #[allow(clippy::type_complexity)]
+    fn knn_points(
+        &self,
+        query: &Self::Point,
+        num: usize,
+    ) -> Vec<(
+        Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>,
+        &'_ Self::Point,
+    )> {
         self.knn(query, num)
             .into_iter()
             .map(|n| (n, self.get_point(n.index)))
@@ -118,7 +127,14 @@ where
     ///
     /// For many KNN search algorithms, the returned neighbors are approximate, and may not
     /// be the actual nearest neighbors.
-    fn nn_point(&self, query: &P) -> Option<(Neighbor<P::Metric, Ix>, &'_ P)> {
+    #[allow(clippy::type_complexity)]
+    fn nn_point(
+        &self,
+        query: &Self::Point,
+    ) -> Option<(
+        Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>,
+        &'_ Self::Point,
+    )> {
         self.nn(query).map(|n| (n, self.get_point(n.index)))
     }
 }
@@ -128,23 +144,29 @@ where
 /// This is not the final API. Eventually, the iterator type will be chosen by the collection,
 /// but for now it is a [`Vec`] until Rust stabilizes GATs.
 #[cfg(feature = "alloc")]
-pub trait KnnMap<K, V, Ix = usize>: KnnPoints<K, Ix>
-where
-    K: MetricPoint,
-    Ix: Copy,
-{
+pub trait KnnMap: KnnPoints {
+    type Value;
+
     /// Get a value using a neighbor index returned by [`Knn::knn`] or [`Knn::nn`].
     ///
     /// This should only be used directly after one of the mentioned methods are called to retrieve
     /// a value associated with a neighbor, and will panic if the index is incorrect due to
     /// mutating the data structure thereafter. The index is only valid up until the next mutation.
-    fn get_value(&self, index: Ix) -> &'_ V;
+    fn get_value(&self, index: Self::Ix) -> &'_ Self::Value;
 
     /// Get `num` nearest neighbor keys of `target`.
     ///
     /// For many KNN search algorithms, the returned neighbors are approximate, and may not
     /// be the actual nearest neighbors.
-    fn knn_values(&self, query: &K, num: usize) -> Vec<(Neighbor<K::Metric, Ix>, &'_ V)> {
+    #[allow(clippy::type_complexity)]
+    fn knn_values(
+        &self,
+        query: &Self::Point,
+        num: usize,
+    ) -> Vec<(
+        Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>,
+        &'_ Self::Value,
+    )> {
         self.knn(query, num)
             .into_iter()
             .map(|n| (n, self.get_value(n.index)))
@@ -155,7 +177,14 @@ where
     ///
     /// For many KNN search algorithms, the returned neighbors are approximate, and may not
     /// be the actual nearest neighbors.
-    fn nn_value(&self, query: &K) -> Option<(Neighbor<K::Metric, Ix>, &'_ V)> {
+    #[allow(clippy::type_complexity)]
+    fn nn_value(
+        &self,
+        query: &Self::Point,
+    ) -> Option<(
+        Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>,
+        &'_ Self::Value,
+    )> {
         self.nn(query).map(|n| (n, self.get_value(n.index)))
     }
 
@@ -166,9 +195,13 @@ where
     #[allow(clippy::type_complexity)]
     fn knn_keys_values(
         &self,
-        query: &K,
+        query: &Self::Point,
         num: usize,
-    ) -> Vec<(Neighbor<K::Metric, Ix>, &'_ K, &'_ V)> {
+    ) -> Vec<(
+        Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>,
+        &'_ Self::Point,
+        &'_ Self::Value,
+    )> {
         self.knn(query, num)
             .into_iter()
             .map(|n| (n, self.get_point(n.index), self.get_value(n.index)))
@@ -180,42 +213,60 @@ where
     /// For many KNN search algorithms, the returned neighbors are approximate, and may not
     /// be the actual nearest neighbors.
     #[allow(clippy::type_complexity)]
-    fn nn_key_value(&self, query: &K) -> Option<(Neighbor<K::Metric, Ix>, &'_ K, &'_ V)> {
+    fn nn_key_value(
+        &self,
+        query: &Self::Point,
+    ) -> Option<(
+        Neighbor<<Self::Metric as Metric<Self::Point>>::Unit, Self::Ix>,
+        &'_ Self::Point,
+        &'_ Self::Value,
+    )> {
         self.nn(query)
             .map(|n| (n, self.get_point(n.index), self.get_value(n.index)))
     }
 }
 
+/// Implement this trait on KNN search data structures that map keys to values and which you can
+/// insert new (key, value) pairs.
+#[cfg(feature = "alloc")]
+pub trait KnnInsert: KnnMap {
+    /// Insert a (key, value) pair to the [`KnnMap`].
+    fn knn(&mut self, key: Self::Point, value: Self::Value);
+}
+
 /// Performs a linear knn search by iterating over everything in the space
 /// and performing a binary search on running set of neighbors.
 ///
 /// ## Example
 ///
 /// ```
-/// use space::{Knn, LinearKnn, MetricPoint, Neighbor};
+/// use space::{Knn, LinearKnn, Metric, Neighbor};
 ///
-/// #[derive(PartialEq)]
-/// struct Hamming(u8);
+/// struct Hamming;
 ///
-/// impl MetricPoint for Hamming {
-///     type Metric = u8;
+/// impl Metric<u8> for Hamming {
+///     type Unit = u8;
 ///
-///     fn distance(&self, other: &Self) -> Self::Metric {
-///         (self.0 ^ other.0).count_ones() as u8
+///     fn distance(&self, &a: &u8, &b: &u8) -> Self::Unit {
+///         (a ^ b).count_ones() as u8
 ///     }
 /// }
 ///
 /// let data = [
-///     Hamming(0b1010_1010),
-///     Hamming(0b1111_1111),
-///     Hamming(0b0000_0000),
-///     Hamming(0b1111_0000),
-///     Hamming(0b0000_1111),
+///     0b1010_1010,
+///     0b1111_1111,
+///     0b0000_0000,
+///     0b1111_0000,
+///     0b0000_1111,
 /// ];
 ///
-/// let search = LinearKnn(data.iter());
+/// let search = LinearKnn {
+///     metric: Hamming,
+///     iter: data.iter(),
+/// };
+///
 /// assert_eq!(
-///     &search.knn(&Hamming(0b0101_0000), 3),
+///     &search.knn(&0b0101_0000, 3),
 ///     &[
 ///         Neighbor { index: 2, distance: 2 },
 ///         Neighbor { index: 3, distance: 2 },
@@ -224,22 +275,28 @@ where
 /// );
 /// ```
 #[cfg(feature = "alloc")]
-pub struct LinearKnn<I>(pub I);
+pub struct LinearKnn<M, I> {
+    pub metric: M,
+    pub iter: I,
+}
 
 #[cfg(feature = "alloc")]
-impl<'a, P: 'a, I> Knn<P> for LinearKnn<I>
+impl<'a, M, P: 'a, I> Knn for LinearKnn<M, I>
 where
-    P: MetricPoint,
+    M: Metric<P>,
     I: Clone + Iterator<Item = &'a P>,
 {
-    type KnnIter = Vec<Neighbor<P::Metric>>;
+    type Ix = usize;
+    type Metric = M;
+    type Point = P;
+    type KnnIter = Vec<Neighbor<M::Unit>>;
 
     fn knn(&self, query: &P, num: usize) -> Self::KnnIter {
         // Create an iterator mapping the dataset into `Neighbor`.
         let mut dataset = self
-            .0
+            .iter
             .clone()
-            .map(|point| point.distance(query))
+            .map(|point| self.metric.distance(point, query))
             .enumerate()
             .map(|(index, distance)| Neighbor { index, distance });
 
@@ -266,11 +323,11 @@ where
         neighbors
     }
 
-    fn nn(&self, query: &P) -> Option<Neighbor<P::Metric>> {
+    fn nn(&self, query: &P) -> Option<Neighbor<M::Unit>> {
         // Map the input iterator into neighbors and then find the smallest one by distance.
-        self.0
+        self.iter
             .clone()
-            .map(|point| point.distance(query))
+            .map(|point| self.metric.distance(point, query))
             .enumerate()
             .map(|(index, distance)| Neighbor { index, distance })
             .min_by_key(|n| n.distance)
